from .addPriorCount import addPriorCount
from .adjustedProfileLik import adjustedProfileLik
from .aveLogCPM import aveLogCPM
from .binomTest import binomTest
from .DGEGLM import DGEGLM
from .DGEList import DGEList
from .dispCoxReid import dispCoxReid
from .dispCoxReidInterpolateTagwise import dispCoxReidInterpolateTagwise
from .estimateGLMCommonDisp import estimateGLMCommonDisp
from .estimateGLMTagwiseDisp import estimateGLMTagwiseDisp
from .exactTest import exactTest
from .exactTestBetaApprox import exactTestBetaApprox
from .exactTestByDeviance import exactTestByDeviance
from .exactTestBySmallP import exactTestBySmallP
from .exactTestDoubleTail import exactTestDoubleTail
from .glmFit import glmFit, glmLRT
from .glmQLFit import glmQLFit, glmQLFTest, plotQLDisp
from .maximizeInterpolant import maximizeInterpolant
from .mglmLevenberg import mglmLevenberg
from .mglmOneGroup import mglmOneGroup
from .mglmOneWay import mglmOneWay, designAsFactor
from .movingAverageByCol import movingAverageByCol
from .nbinomDeviance import nbinomDeviance
from .predFC import predFC
from .q2qnbinom import q2qnbinom
from .splitIntoGroups import splitIntoGroups
from .stats import pnbinom, qnbinom
from .systematicSubset import systematicSubset
from .topTags import topTags
from .validDGEList import validDGEList
from .edgepy_cpp import *
